==> 1.4.9.tar.gz: lib/src/fragment/fragment_container.dart <==
import 'dart:collection';
import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutterx_application/flutterx_application.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';
import 'package:flutterx_preferences/flutterx_preferences.dart';
import 'package:flutterx_utils/flutterx_utils.dart';

part 'fragment.dart';
part 'fragment_manager.dart';

class FragmentContainer extends StatefulWidget {
  final AxisDirection axisDirection;
  final bool primary;
  final Storage? storage;
  final FragmentEntry? initialFragment;
  final Widget Function(Fragment fragment) fragmentWrapper;
  final List<Fragment> fragments;

  @override
  FragmentContainerKey get key => super.key as FragmentContainerKey;

  const FragmentContainer({
    required FragmentContainerKey key,
    this.axisDirection = AxisDirection.right,
    this.primary = true,
    this.storage,
    this.initialFragment,
    this.fragmentWrapper = removePrimaryScrollController,
    required this.fragments,
  })  : assert(fragments.length > 0, 'fragments cannot be empty'),
        super(key: key);

  @override
  State<FragmentContainer> createState() => FragmentContainerState();

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(EnumProperty<AxisDirection>('axisDirection', axisDirection))
      ..add(DiagnosticsProperty<bool>('primary', primary))
      ..add(DiagnosticsProperty<Storage?>('storage', storage))
      ..add(ObjectFlagProperty<Widget Function(Fragment)>.has('fragmentWrapper', fragmentWrapper))
      ..add(DiagnosticsProperty<FragmentEntry?>('initialFragment', initialFragment));
  }

  static Widget removePrimaryScrollController(Fragment fragment) => PrimaryScrollController.none(child: fragment);

  static FragmentManager of(BuildContext context) => context.findAncestorStateOfType<FragmentContainerState>()!;

  static FragmentManager ofName(String name) => FragmentContainerKey(name).currentState!;
}

class FragmentContainerState extends State<FragmentContainer> implements FragmentManager {
  late PageController _controller;
  List<Fragment> _fragments = const [];
  List<Widget> _children = const [];
  Map<FragmentKey, int> _indexByKey = const {};
  bool _forward = true;

  @override
  FragmentContainer get container => widget;

  @override
  late FragmentStack currentFragment;

  @override
  LiveData<bool> get canGoBack => currentFragment.canUndo;

  @override
  LiveData<bool> get canGoForward => currentFragment.canRedo;

  @override
  void initState() {
    super.initState();
    _indexFragments();
    _initialize();
  }

  @override
  void didUpdateWidget(covariant FragmentContainer oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (_fragments != widget.fragments) _indexFragments();
    if (oldWidget.key != widget.key || !identical(oldWidget.storage, widget.storage)) _initialize(update: true);
  }

  @override
  Widget build(BuildContext context) {
    if (!currentFragment.initialized) {
      currentFragment.value = _initialKey(context);
      _initController();
    }
    return Scrollable(
        key: ValueKey(widget.key),
        axisDirection: widget.axisDirection,
        controller: _controller,
        physics: const NeverScrollableScrollPhysics(),
        restorationId: widget.key.name,
        scrollBehavior: ScrollConfiguration.of(context).copyWith(scrollbars: false),
        viewportBuilder: (context, position) =>
            Viewport(cacheExtent: 0, axisDirection: widget.axisDirection, offset: position, slivers: <Widget>[
              SliverFillViewport(
                  viewportFraction: _controller.viewportFraction, delegate: SliverChildListDelegate(_children)),
            ]));
  }

  @override
  void navigateTo<T extends Fragment>({bool forward = true, dynamic args}) {
    for (final item in _fragments) {
      if (item is T) {
        _forward = forward;
        currentFragment.value = FragmentEntry(item.key, args != null ? jsonEncode(args) : null);
        return;
      }
    }
    throw StateError('No fragment found with type $T');
  }

  @override
  void navigateToKey(FragmentKey key, {bool forward = true, dynamic args}) {
    _forward = forward;
    currentFragment.value = FragmentEntry(key, args != null ? jsonEncode(args) : null);
  }

  @override
  bool back({bool forward = false}) {
    if (currentFragment.canUndo.value) {
      _forward = forward;
      currentFragment.historyUndo();
      return true;
    }
    return false;
  }

  @override
  bool forward({bool forward = true}) {
    if (currentFragment.canRedo.value) {
      _forward = forward;
      currentFragment.historyRedo();
      return true;
    }
    return false;
  }

  void _indexFragments() {
    _fragments = widget.fragments;
    _children = KeyedSubtree.ensureUniqueKeysForList(_fragments.map(widget.fragmentWrapper).toList(growable: false));
    var i = 0;
    _indexByKey = {for (final fragment in _fragments) fragment.key: i++};
  }

  void _initialize({bool update = false}) {
    if (update) currentFragment.removeObserver(_pushEntry);
    currentFragment = FragmentStack._(key: '${widget.key.name}_active_fragment', storage: widget.storage)
      ..observeForever(_pushEntry, dispatcher: Dispatcher.postFrame);
    if (currentFragment.initialized) _initController();
  }

  Future<void> _pushEntry(FragmentEntry entry) async {
    final currentKey = entry.key;
    final lastIndex = _controller.page!.round();
    final currentIndex = _indexByKey[currentKey];
    if (currentIndex == lastIndex) return;
    if (currentIndex == null) {
      currentFragment.value = _initialKey(context);
      return;
    }
    final lastKey = _indexByKey.keys.elementAt(lastIndex);
    lastKey.currentState?.let((state) => state._active ? state.onPause() : null);
    currentKey.currentState?.setArguments(entry.args);
    if ((_forward && currentIndex - lastIndex == 1) || (!_forward && currentIndex - lastIndex == -1)) {
      await _controller.animateToPage(currentIndex, duration: kThemeAnimationDuration, curve: Curves.ease);
    } else {
      final startIndex = _forward
          ? currentIndex == 0
              ? 0
              : currentIndex - 1
          : currentIndex == _children.length - 1
              ? currentIndex
              : currentIndex + 1;
      final endIndex = _forward
          ? currentIndex == 0
              ? 1
              : currentIndex
          : currentIndex == _children.length - 1
              ? currentIndex - 1
              : currentIndex;
      final originalChildren = _children;
      setState(() {
        _children = List<Widget>.of(_children, growable: false);
        final tmp = _children[startIndex];
        _children[startIndex] = _children[lastIndex];
        _children[lastIndex] = tmp;
      });
      _controller.jumpToPage(startIndex);
      await _controller.animateToPage(endIndex, duration: kThemeAnimationDuration, curve: Curves.ease);
      if (mounted) {
        setState(() => widget.fragments != _fragments ? _indexFragments() : _children = originalChildren);
        if (currentIndex != endIndex) _controller.jumpToPage(currentIndex);
      }
    }
    if (!mounted) return;
    if (widget.primary) AppPage.of(context).fragment = currentKey;
    currentKey.currentState?.let((state) => state._active ? null : state.onResume());
  }

  FragmentEntry _initialKey(BuildContext context) {
    if (widget.primary) {
      final key = AppPage.of(context).fragment;
      if (key != null && _indexByKey.containsKey(key)) return FragmentEntry(key);
    }
    final entry = widget.initialFragment;
    if (entry != null && _indexByKey.containsKey(entry)) return entry;
    return FragmentEntry(_fragments.first.key);
  }

  void _initController() =>
      _controller = PageController(initialPage: _indexByKey[currentFragment.value.key]!, keepPage: false);

  void _onFragmentAttached(FragmentState state) {
    final current = currentFragment.value;
    if (current.key == state.widget.key) {
      state.setArguments(current.args);
      if (!state._active) state.onResume();
    }
  }

  void _onFragmentDetached(FragmentState state) {
    final current = currentFragment.value;
    if (current.key == state.widget.key) if (state._active) state.onPause();
  }
}

@immutable
class FragmentEntry with DTO {
  final FragmentKey key;
  final String? args;

  const FragmentEntry(this.key, [this.args]);

  factory FragmentEntry.fromJson(JsonObject json) => FragmentEntry(FragmentKey(json['name']), json['args']);

  @override
  JsonObject toJson() => {'name': key.name, if (args != null) 'args': args};

  @override
  bool operator ==(Object other) => other is FragmentEntry && other.key == key && other.args == args;

  @override
  int get hashCode => hashValues(key, args);
}

class FragmentStack extends MutableLiveData<FragmentEntry> with HistoryMixin<FragmentEntry> {
  final String _indexKey;
  final Storage? _storage;
  final Queue<String>? _queue;

  @override
  Queue<String> get queue => _queue ?? super.queue;

  FragmentStack._({required String key, required Storage? storage})
      : _indexKey = '${key}_index',
        _storage = storage,
        _queue = storage == null ? null : StorageLiveQueue(key: key, storage: storage),
        super.late() {
    if (storage != null) {
      final index = storage[_indexKey]?.let(int.parse) ?? -1;
      if (index != -1) {
        historyIndex = index;
        restoreState();
      }
    }
  }

  @override
  void updateValues() {
    super.updateValues();
    final storage = _storage;
    if (storage != null) storage[_indexKey] = historyIndex.toString();
  }

  @override
  String serialize(FragmentEntry data) => jsonEncode(data.toJson());

  @override
  FragmentEntry deserialize(String data) => FragmentEntry.fromJson(jsonDecode(data));
}

class FragmentContainerKey extends GlobalKey<FragmentContainerState> {
  final String name;

  const FragmentContainerKey(this.name) : super.constructor();

  @override
  bool operator ==(Object other) => other is FragmentContainerKey && identical(other.name, name);

  @override
  int get hashCode => identityHashCode(name);

  @override
  String toString() => 'FragmentContainerKey($name)';
}

==> 1.5.1.tar.gz: lib/src/fragment/fragment_container.dart <==
import 'dart:collection';
import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutterx_application/flutterx_application.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';
import 'package:flutterx_preferences/flutterx_preferences.dart';
import 'package:flutterx_utils/flutterx_utils.dart';

part 'fragment.dart';
part 'fragment_manager.dart';

class FragmentContainer extends StatefulWidget {
  final AxisDirection axisDirection;
  final bool primary;
  final Storage? storage;
  final FragmentEntry? initialFragment;
  final Widget Function(Fragment fragment) fragmentWrapper;
  final List<Fragment> fragments;

  @override
  FragmentContainerKey get key => super.key as FragmentContainerKey;

  const FragmentContainer({
    required FragmentContainerKey key,
    this.axisDirection = AxisDirection.right,
    this.primary = true,
    this.storage,
    this.initialFragment,
    this.fragmentWrapper = removePrimaryScrollController,
    required this.fragments,
  })  : assert(fragments.length > 0, 'fragments cannot be empty'),
        super(key: key);

  @override
  State<FragmentContainer> createState() => FragmentContainerState();

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(EnumProperty<AxisDirection>('axisDirection', axisDirection))
      ..add(DiagnosticsProperty<bool>('primary', primary))
      ..add(DiagnosticsProperty<Storage?>('storage', storage))
      ..add(ObjectFlagProperty<Widget Function(Fragment)>.has('fragmentWrapper', fragmentWrapper))
      ..add(DiagnosticsProperty<FragmentEntry?>('initialFragment', initialFragment));
  }

  static Widget removePrimaryScrollController(Fragment fragment) => PrimaryScrollController.none(child: fragment);

  static FragmentManager of(BuildContext context) => context.findAncestorStateOfType<FragmentContainerState>()!;

  static FragmentManager ofName(String name) => FragmentContainerKey(name).currentState!;
}

class FragmentContainerState extends State<FragmentContainer> implements FragmentManager {
  late PageController _controller;
  List<Fragment> _fragments = const [];
  List<Widget> _children = const [];
  Map<FragmentKey, int> _indexByKey = const {};
  ScrollDirection _direction = ScrollDirection.idle;

  @override
  FragmentContainer get container => widget;

  @override
  late FragmentStack currentFragment;

  @override
  LiveData<bool> get canGoBack => currentFragment.canUndo;

  @override
  LiveData<bool> get canGoForward => currentFragment.canRedo;

  @override
  void initState() {
    super.initState();
    _indexFragments();
    _initialize();
  }

  @override
  void didUpdateWidget(covariant FragmentContainer oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (_fragments != widget.fragments) _indexFragments();
    if (oldWidget.key != widget.key || !identical(oldWidget.storage, widget.storage)) _initialize(update: true);
  }

  @override
  Widget build(BuildContext context) {
    if (!currentFragment.initialized) {
      currentFragment.value = _initialKey(context);
      _initController();
    }
    return Scrollable(
        key: ValueKey(widget.key),
        axisDirection: widget.axisDirection,
        controller: _controller,
        physics: const NeverScrollableScrollPhysics(),
        restorationId: widget.key.name,
        scrollBehavior: ScrollConfiguration.of(context).copyWith(scrollbars: false),
        viewportBuilder: (context, position) =>
            Viewport(cacheExtent: 0, axisDirection: widget.axisDirection, offset: position, slivers: <Widget>[
              SliverFillViewport(
                  viewportFraction: _controller.viewportFraction, delegate: SliverChildListDelegate(_children)),
            ]));
  }

  @override
  void navigateTo<T extends Fragment>({ScrollDirection direction = ScrollDirection.idle, dynamic args}) {
    for (final item in _fragments) {
      if (item is T) {
        _direction = direction;
        currentFragment.value = FragmentEntry(item.key, args != null ? jsonEncode(args) : null);
        return;
      }
    }
    throw StateError('No fragment found with type $T');
  }

  @override
  void navigateToKey(FragmentKey key, {ScrollDirection direction = ScrollDirection.idle, dynamic args}) {
    _direction = direction;
    currentFragment.value = FragmentEntry(key, args != null ? jsonEncode(args) : null);
  }

  @override
  bool back({ScrollDirection direction = ScrollDirection.reverse}) {
    if (currentFragment.canUndo.value) {
      _direction = direction;
      currentFragment.historyUndo();
      return true;
    }
    return false;
  }

  @override
  bool forward({ScrollDirection direction = ScrollDirection.forward}) {
    if (currentFragment.canRedo.value) {
      _direction = direction;
      currentFragment.historyRedo();
      return true;
    }
    return false;
  }

  void _indexFragments() {
    _fragments = widget.fragments;
    _children = KeyedSubtree.ensureUniqueKeysForList(_fragments.map(widget.fragmentWrapper).toList(growable: false));
    var i = 0;
    _indexByKey = {for (final fragment in _fragments) fragment.key: i++};
  }

  void _initialize({bool update = false}) {
    if (update) currentFragment.removeObserver(_pushEntry);
    currentFragment = FragmentStack._(key: '${widget.key.name}_active_fragment', storage: widget.storage)
      ..observeForever(_pushEntry, dispatcher: Dispatcher.postFrame);
    if (currentFragment.initialized) _initController();
  }

  Future<void> _pushEntry(FragmentEntry entry) async {
    final currentKey = entry.key;
    final lastIndex = _controller.page!.round();
    final currentIndex = _indexByKey[currentKey];
    if (currentIndex == lastIndex) return;
    if (currentIndex == null) {
      currentFragment.value = _initialKey(context);
      return;
    }
    final lastKey = _indexByKey.keys.elementAt(lastIndex);
    lastKey.currentState?.let((state) => state._active ? state.onPause() : null);
    currentKey.currentState?.setArguments(entry.args);
    final bool forward;
    switch (_direction) {
      case ScrollDirection.idle:
        forward = lastIndex < currentIndex;
        break;
      case ScrollDirection.forward:
        forward = true;
        break;
      case ScrollDirection.reverse:
        forward = false;
        break;
    }
    if ((forward && currentIndex - lastIndex == 1) || (!forward && currentIndex - lastIndex == -1)) {
      await _controller.animateToPage(currentIndex, duration: kThemeAnimationDuration, curve: Curves.ease);
    } else {
      final startIndex = forward
          ? currentIndex == 0
              ? 0
              : currentIndex - 1
          : currentIndex == _children.length - 1
              ? currentIndex
              : currentIndex + 1;
      final endIndex = forward
          ? currentIndex == 0
              ? 1
              : currentIndex
          : currentIndex == _children.length - 1
              ? currentIndex - 1
              : currentIndex;
      final originalChildren = _children;
      setState(() {
        _children = List<Widget>.of(_children, growable: false);
        final tmp = _children[startIndex];
        _children[startIndex] = _children[lastIndex];
        _children[lastIndex] = tmp;
      });
      _controller.jumpToPage(startIndex);
      await _controller.animateToPage(endIndex, duration: kThemeAnimationDuration, curve: Curves.ease);
      if (mounted) {
        setState(() => widget.fragments != _fragments ? _indexFragments() : _children = originalChildren);
        if (currentIndex != endIndex) _controller.jumpToPage(currentIndex);
      }
    }
    if (!mounted) return;
    if (widget.primary) AppPage.of(context).fragment = currentKey;
    currentKey.currentState?.let((state) => state._active ? null : state.onResume());
  }

  FragmentEntry _initialKey(BuildContext context) {
    if (widget.primary) {
      final key = AppPage.of(context).fragment;
      if (key != null && _indexByKey.containsKey(key)) return FragmentEntry(key);
    }
    final entry = widget.initialFragment;
    if (entry != null && _indexByKey.containsKey(entry)) return entry;
    return FragmentEntry(_fragments.first.key);
  }

  void _initController() =>
      _controller = PageController(initialPage: _indexByKey[currentFragment.value.key]!, keepPage: false);

  void _onFragmentAttached(FragmentState state) {
    final current = currentFragment.value;
    if (current.key == state.widget.key) {
      state.setArguments(current.args);
      if (!state._active) state.onResume();
    }
  }

  void _onFragmentDetached(FragmentState state) {
    final current = currentFragment.value;
    if (current.key == state.widget.key) if (state._active) state.onPause();
  }
}

@immutable
class FragmentEntry with DTO {
  final FragmentKey key;
  final String? args;

  const FragmentEntry(this.key, [this.args]);

  factory FragmentEntry.fromJson(JsonObject json) => FragmentEntry(FragmentKey(json['name']), json['args']);

  @override
  JsonObject toJson() => {'name': key.name, if (args != null) 'args': args};

  @override
  bool operator ==(Object other) => other is FragmentEntry && other.key == key && other.args == args;

  @override
  int get hashCode => hashValues(key, args);
}

class FragmentStack extends MutableLiveData<FragmentEntry> with HistoryMixin<FragmentEntry> {
  final String _indexKey;
  final Storage? _storage;
  final Queue<String>? _queue;

  @override
  Queue<String> get queue => _queue ?? super.queue;

  FragmentStack._({required String key, required Storage? storage})
      : _indexKey = '${key}_index',
        _storage = storage,
        _queue = storage == null ? null : StorageLiveQueue(key: key, storage: storage),
        super.late() {
    if (storage != null) {
      final index = storage[_indexKey]?.let(int.parse) ?? -1;
      if (index != -1) {
        historyIndex = index;
        restoreState();
      }
    }
  }

  @override
  void updateValues() {
    super.updateValues();
    final storage = _storage;
    if (storage != null) storage[_indexKey] = historyIndex.toString();
  }

  @override
  String serialize(FragmentEntry data) => jsonEncode(data.toJson());

  @override
  FragmentEntry deserialize(String data) => FragmentEntry.fromJson(jsonDecode(data));
}

class FragmentContainerKey extends GlobalKey<FragmentContainerState> {
  final String name;

  const FragmentContainerKey(this.name) : super.constructor();

  @override
  bool operator ==(Object other) => other is FragmentContainerKey && identical(other.name, name);

  @override
  int get hashCode => identityHashCode(name);

  @override
  String toString() => 'FragmentContainerKey($name)';
}
